--[[
Name: targetAssist
Author: Sonora (The Dragonflight, US Kirin Tor)
Website & SVN: http://wow.curseforge.com/addons/targetassist/
License: GNU General Public License v3
]]

--Ugly global namespace bindings for binding.xml
BINDING_HEADER_targetAssist = "targetAssist"
BINDING_NAME_TAADDTARGET = "Add current target"
BINDING_NAME_TACLEARTARGETS = "Clear all targets"

--Utilities
local function deepcopy(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, _copy(getmetatable(object)))
    end
    return _copy(object)
end

local function hasValue(t, value)
    for _, v in ipairs(t) do
        if v == value then return true end
    end
    for k, v in pairs(t) do
        if v == value then return true end
    end
end
local function keys(t)
    ks = {}
    for k,_ in pairs(t) do table.insert(ks, k) end
    return ks
end
local function keyFromValue(t, value)
    for k,v in pairs(t) do
        if v == value then return k end
    end
end
local function tableIndex(t, value)
    for idx, v in ipairs(t) do
        if v == value then return idx end
    end
end


--Ace3 addon application object & Libraries
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")

targetAssist = LibStub("AceAddon-3.0"):NewAddon("targetAssist", "AceConsole-3.0", "AceComm-3.0", "AceSerializer-3.0", "AceEvent-3.0", "AceTimer-3.0")

function targetAssist:OnInitialize()
    
    --Addon wide data structures
    self.roster = {} --{unitName=unitID,}
    self.mainassists = {}
    self.tanks = {}
    
    self.buttonWidth = 0
    self.postCombatCalls = {}
    self.previousBroadcasts = {} --{{fromPlayer, channel, targets},}
    
    --Setup a default profile table for the database
    local defaultProfileDb = {
        profile = {
            config = deepcopy(self.defaultConfig)
        }
    }
    
    --Setup database communication    
    self.db = LibStub("AceDB-3.0"):New("targetAssistDb", defaultProfileDb)
    self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
    self.db.RegisterCallback(self, 'OnNewProfile', 'OnNewProfile')
    
    --Register us as recievers of our AceComm messages
    self:RegisterComm('ta.bT', 'targetReceiver')
    
    --If SharedMedia is loaded, share our media
    if IsAddOnLoaded("SharedMedia") then
        local lib = LibStub("LibSharedMedia-3.0")
        lib:Register(lib.MediaType.STATUSBAR, "targetAssist.Bar", [=[Interface\Addons\targetAssist\media\barTex]=])
    end

    --This call in turn handles changes to targetAssist.config and config GUI construction
    self:OnProfileChanged(nil, self.db, self.db:GetCurrentProfile())
    
    --Two variables determine whether or not the addon is shown & events are registered
    self.hide = false --managed by addon based on whether or not the player is in a group and we auto-hide out of group
    
    --Event registration
    self:RegisterEvent('PARTY_MEMBERS_CHANGED')
    self:RegisterEvent('RAID_ROSTER_UPDATE')
    self:RegisterEvent('UNIT_PET')
    self:RegisterEvent('PLAYER_ROLES_ASSIGNED')
    self:RegisterEvent("oRA_MainTankUpdate")
    
    --Setup timers
    self:ScheduleRepeatingTimer('onUpdate', self.config.updateInterval)

    --oRA3 event notification
    if oRA3 and oRA3.GetSortedTanks then
        oRA3.RegisterCallback(targetAssist, "oRA3_TanksUpdated")
        targetAssist:oRA3_TanksUpdated(nil, oRA3:GetSortedTanks())
    end
    
    --Update our shown state
    self:updateEnabled()
    
    --Setup the user interface, and fire events which don't on their own at log in
    self:buildUI()
    
end

function targetAssist:__updateFrameState(stateCall)
    return function()
        if self.headerFrame then
            self.headerFrame[stateCall](self.headerFrame)
            for _, buttonFrame in pairs(self.targetButtons or {}) do
                buttonFrame[stateCall](buttonFrame)
            end
        end
    end
end
function targetAssist:updateEnabled()

    if (self.hide or self.config.userDisabled) then
        self:registerPostCombatCall(self:__updateFrameState('Hide'))
        self.enabled = false
    else        
        self:registerPostCombatCall(self:__updateFrameState('Show'))
        self.enabled = true
    end
end

function targetAssist:buildUI()
    
    --Setup options panels
    AceConfig:RegisterOptionsTable('targetAssist', self:createBasePanel())
    self.optionsBaseFrame = AceConfigDialog:AddToBlizOptions("targetAssist", "targetAssist")
    
    AceConfig:RegisterOptionsTable("targetAssist.Features", self:createFeaturesPanel())
    AceConfigDialog:AddToBlizOptions("targetAssist.Features", "Features", "targetAssist")
    
    AceConfig:RegisterOptionsTable("targetAssist.Appearance", self:createAppearancePanel())
    AceConfigDialog:AddToBlizOptions("targetAssist.Appearance", "Appearance", "targetAssist")
    
    AceConfig:RegisterOptionsTable("targetAssist.Profiles", LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db))
    AceConfigDialog:AddToBlizOptions("targetAssist.Profiles", "Profiles", "targetAssist")
    
    --Setup command line options
    AceConfig:RegisterOptionsTable("targetAssist.Console", self:createConsoleOptions(), {"targetAssist", "ta"})
    
    --Build a context menu
    self.contextMenu = CreateFrame("Frame", "targetAssistContextMenu", UIParent, "UIDropDownMenuTemplate")
    
end

function targetAssist:targetReceiver(prefix, message, distribution, sender)

    if not UnitIsUnit('player',sender) then

        local success, newTargets = self:Deserialize(message)
    
        if #self.previousBroadcasts == self.config.broadcastHistory then
            table.remove(self.previousBroadcasts, 1)
        end
        table.insert(self.previousBroadcasts, {sender, distribution, newTargets})
    
        if self.config.autoAcceptBroadcasts then
            self.config.targets = newTargets
            self:updateConfig()
        else
            self:yesnoBox('A set of assist targets has been sent by '..sender..'.  Would you like to change to this new target list?', 
                function()
                    self.config.targets = newTargets
                    self:updateConfig()
                end
            )
        end
    
    end
end

function targetAssist:upgradeDatabase(config)

    if config.dbVersion == self.currentDbVersion then return config
    else
        local nextVersion = config.dbVersion + 1
        local migrationCall = self.migrationPaths[nextVersion]
        
        if migrationCall then migrationCall(config) end
        
        config.dbVersion = nextVersion
        return self:upgradeDatabase(config)
    end
    
end

function targetAssist:OnProfileChanged(eventName, db, newProfile)
    
    self.config = db.profile.config

    if self.config then
    
        --Unversioned databases are set to v. 1
        if self.config.dbVersion == nil then self.config.dbVersion = 1 end

        --Handle database version checking and upgrading if necessary
        local startVersion = self.config.dbVersion
        self:upgradeDatabase(self.config)
        if startVersion ~= self.config.dbVersion then
            print("targetAssist configuration database upgraded from v."..startVersion.." to v."..self.config.dbVersion)
        end
    
    end
    
    self:updateRoster()
    self:updateConfig()
    
end

function targetAssist:OnNewProfile(eventName, db, profile)

    --Set the dbVersion to the most recent, as defaults for the new profile should be up-to-date
    db.profile.config.dbVersion = self.currentDbVersion
    
end

function targetAssist:updateConfig() self.updateConfigPostCombat = true end

function targetAssist:__updateConfig()
    self.updateConfigPostCombat = false
    self:createHeaderFrame()
    self:createTargetButtons()
end

function targetAssist:playerInGroup()
    return GetNumGroupMembers() > 0 or GetNumGroupMembers() > 0
end

function targetAssist:addTanks()

    table.sort(self.tanks)
    for i, tankName in ipairs(self.tanks) do
        if not hasValue(self.config.targets, tankName) then
            table.insert(self.config.targets, i, tankName)
        end
    end
    self:updateConfig()

end

function targetAssist:addMainAssists()
    
    table.sort(self.mainassists)
    for i, assistName in ipairs(self.mainassists) do
        if not hasValue(self.config.targets, assistName) then
            table.insert(self.config.targets, i, assistName)
        end
    end
    self:updateConfig()
    
end
function targetAssist:addTarget()
    
    local unitName = GetUnitName('target')
    if unitName then
        if not UnitIsFriend('player', 'target') then
            unitName = GetUnitName('targettarget')
        end
        if not hasValue(self.config.targets, unitName) then
            table.insert(self.config.targets, 1, unitName)
        end
    end
    self:updateConfig()
    
end
function targetAssist:addAll()
    local unitNames = keys(self.roster)
    for _, unitName in ipairs(unitNames) do
        if not hasValue(self.config.targets, unitName) then
            table.insert(self.config.targets, unitName)
        end
    end
    self:updateConfig()
end
function targetAssist:addParty()
    for i=1,GetNumGroupMembers() do
        local unitName = UnitName('party'..i)
        if unitName and not hasValue(self.config.targets, unitName) then
            table.insert(self.config.targets, unitName)
        end
    end
    self:updateConfig()
end
function targetAssist:clearAll()
    self.config.targets = {}
    self:updateConfig()
end
function targetAssist:invertTargetOrder()
    self.config.invertTargetOrder = not self.config.invertTargetOrder
    self:updateConfig()
end

function targetAssist:updateRoster()
    local autoAdds = {}
    for _, v in ipairs(self.mainassists) do autoAdds[v] = false end
    for _, v in ipairs(self.tanks) do autoAdds[v] = false end

    self.mainassists = {}
    self.tanks = {}
    
    self.roster = {}  --{unitName = unitID,}
    
    if self.config.showPets then
        local petName = GetUnitName('pet')
        if petName then
            self.roster[petName..' (your pet)'] = 'pet'
        end
    end
    
    if self.config.includePlayer then
        self.roster[GetUnitName('player')] = 'player'
    end
    
    if self:playerInGroup() then    
        for i=1,GetNumGroupMembers() do
            local unitName = GetUnitName('party'..i)
            if unitName then 
                self.roster[unitName] = 'party'..i 
                local role = UnitGroupRolesAssigned('party'..i)
                if role == 'TANK' then table.insert(self.tanks, unitName) end
        
                if self.config.showPets then
                    local petName = GetUnitName('party'..i..'pet')
                    if petName then
                        self.roster[petName..' ('..unitName.."'s pet)"] = 'party'..i..'pet'
                    end
                end
            end
        end
        for i=1,GetNumGroupMembers() do
            local unitName = GetUnitName('raid'..i)
            if unitName and not UnitIsUnit('player','raid'..i) then 
                self.roster[unitName] = 'raid'..i
                local name, _, _, _, _, _, _, _, _, role = GetRaidRosterInfo(i)
                if role == 'MAINTANK' then table.insert(self.tanks, unitName) end
                if role == 'MAINASSIST' then table.insert(self.mainassists, unitName) end
        
                if self.config.showPets then
                    local petName = GetUnitName('raid'..i..'pet')
                    if petName then
                        self.roster[petName..' ('..unitName.."'s pet)"] = 'raid'..i..'pet'
                    end
                end
        
            end
        end
    
        if oRA and oRA.maintanktable and self.config.autoUpdateORA2 then
            for _, unitName in ipairs(oRA.maintanktable) do
                table.insert(self.tanks, unitName)
                if autoAdds[unitName] ~= nil then autoAdds[unitName] = true end
            end
        end
        if oRA3 and self.oRA3Tanks and self.config.autoUpdateORA3 then
            for _, unitName in ipairs(self.oRA3Tanks) do
                table.insert(self.tanks, unitName)
                if autoAdds[unitName] ~= nil then autoAdds[unitName] = true end
            end
        end
        
        --Purge targets we previously auto-added but are now missing
        if self.config.purgeAutoAdds then
            for unitName,currentAdd in pairs(autoAdds) do
                if not currentAdd then
                    table.remove(self.config.targets, tableIndex(self.config.targets, unitName))
                end
            end
        end
        
        if self.config.autoUpdateTanks then self:addTanks() end
        if self.config.autoUpdateMainAssists then self:addMainAssists() end
        
        if self.hide then
            self.hide = false
            self:updateEnabled()
        end
        
    else 
        if self.config.hideOutOfGroup then
            if not self.hide then
                self.hide = true
                self:updateEnabled()
            end
        else
            if self.hide then
                self.hide = false
                self:updateEnabled()
            end
        end
    end
    
    if self.contextMenu and self.contextMenu:IsShown() then self.contextMenu:Hide() end
    
    self:updateConfig()
end

local function getUnitHealth(unitID)
    --returns percentage
    return UnitHealth(unitID) / UnitHealthMax(unitID)
end

function targetAssist:registerPostCombatCall(call) table.insert(self.postCombatCalls, call) end

function targetAssist:checkHostileRange(unitID)
    local _, playerClass = UnitClass('player')
    local spellName = self.config.classRangeChecks[playerClass]
    return IsSpellInRange(self.config.classRangeChecks[playerClass], unitID)
end

local function addUnitID(trackingTable, unitID)
    for foundingTargetID, count in pairs(trackingTable) do
        if UnitIsUnit(unitID, foundingTargetID) then
            trackingTable[foundingTargetID] = trackingTable[foundingTargetID]+1
            return foundingTargetID
        end
    end
    
    trackingTable[unitID] = 1
    return unitID
end
local function getFoundingTargetID(trackingTable, unitID)
    for foundingTargetID, count in pairs(trackingTable) do
        if UnitIsUnit(unitID, foundingTargetID) then
            return foundingTargetID
        end
    end
end
function targetAssist:onUpdate()
--[[
A big ugly function:  keeping everything inline here to help keep track of
of the processing which is happening on each update cylce.
]]--

    if not InCombatLockdown() then
        if self.updateConfigPostCombat then self:__updateConfig() end
    
        for _, call in ipairs(self.postCombatCalls) do call() end
        self.postCombatCalls = {}
    end
    
    if not self.enabled then return false end

    local friendlyTargets, hostileTargets = {}, {} --{foundingTargetID = count,}
    for unitName, unitID in pairs(self.roster) do
        if UnitIsFriend('player', unitID..'target') then
            addUnitID(friendlyTargets, unitID..'target')
        else
            addUnitID(hostileTargets, unitID..'target')
        end
    end
    if UnitIsFriend('player','playertarget') then
        addUnitID(friendlyTargets, 'playertarget')
    else
        addUnitID(hostileTargets, 'playertarget')
    end
    
    local maxCount, maxHostile = 1, nil
    for foundingTargetID, count in pairs(hostileTargets) do
        if count > maxCount then
            maxCount = count
            maxHostile = foundingTargetID
        end
    end
    
    local maxCount, maxFriendly = 1, nil
    for foundingTargetID, count in pairs(friendlyTargets) do
        if count > maxCount then
            maxCount = count
            maxFriendly = foundingTargetID
        end
    end

    for targetName, buttonFrame in pairs(self.targetButtons or {}) do
        local unitName = buttonFrame.unitID and GetUnitName(buttonFrame.unitID)
        if unitName then            
            buttonFrame.unitName = unitName

            local unitHealth = getUnitHealth(buttonFrame.unitID)
            
            --Style the targeted unit's buttons
            if UnitIsUnit('playertarget', buttonFrame.unitID) then
                buttonFrame.texture:SetVertexColor(unpack(self.config.myTargetColor))
                if self.config.showFriendlyHealthPercent then
                    buttonFrame.fontString:SetText(unitName.."\n"..math.floor(unitHealth*100).."%")
                else
                    buttonFrame.fontString:SetText(unitName)
                end
        
            --Style friendly unit buttons
            elseif UnitIsFriend('player', buttonFrame.unitID) then
                local r,g,b,a
                if self.config.colorFriendlyTargetsByClass and UnitIsPlayer(buttonFrame.unitID) then
                    local _, stdClassName = UnitClass(buttonFrame.unitID)
                    local classColor = RAID_CLASS_COLORS[stdClassName]
                    if classColor then
                        r,g,b,a = classColor.r, classColor.g, classColor.b, self.config.friendlyTargetColor[4]
                    else
                        r,g,b,a = unpack(self.config.friendlyTargetColor)
                    end
                else
                    r,g,b,a = unpack(self.config.friendlyTargetColor)
                end
                if self.config.fadeOutOfRange and not UnitInRange(buttonFrame.unitID) then
                    a = max(0.1, a - self.config.outOfRangeAlphaOffset)
                end
                buttonFrame.texture:SetVertexColor(r,g,b,a)
        
                if self.config.showFriendlyHealthPercent then
                    buttonFrame.fontString:SetText(unitName.."\n"..math.floor(unitHealth*100).."%")
                else
                    buttonFrame.fontString:SetText(unitName)
                end
        
            --Style hostile unit buttons
            else
                local r,g,b,a, fontColor
                if self.config.colorHostileTargetsByClass and UnitIsPlayer(buttonFrame.unitID) then
                    local _, stdClassName = UnitClass(buttonFrame.unitID)
                    local classColor = RAID_CLASS_COLORS[stdClassName]
                    r,g,b,a = classColor.r, classColor.g, classColor.b, self.config.hostileTargetColor[4]
                    fontColor = self.config.hostileTargetColor
                else
                    r,g,b,a = unpack(self.config.hostileTargetColor)
                    fontColor = self.config.fontColor
                end
        
                if self.config.fadeOutOfRange and self:checkHostileRange(buttonFrame.unitID) ~= 1 then
                    a = max(0.1, a - self.config.outOfRangeAlphaOffset)
                end
                buttonFrame.texture:SetVertexColor(r,g,b,a)
                buttonFrame.fontString:SetTextColor(unpack(fontColor))
        
                if self.config.showHostileHealthPercent then
                    buttonFrame.fontString:SetText(unitName.."\n"..math.floor(unitHealth*100).."%")
                else
                    buttonFrame.fontString:SetText(unitName)
                end
        
            end
        
            if self.config.showHealth then
                local width = unitHealth * self.buttonWidth
                if width > 0 then
                    buttonFrame.texture:SetWidth( width )
                else
                    buttonFrame.texture:SetWidth( self.buttonWidth )
                    buttonFrame.texture:SetVertexColor(unpack(self.config.missingColor))
                end
            end
        
            if buttonFrame.raidIcon then
                local iconIdx = GetRaidTargetIndex(buttonFrame.unitID)
                if iconIdx then
                    buttonFrame.raidIcon:SetTexture(self.config.raidIcons[iconIdx][2])
                    buttonFrame.raidIcon:Show()
                else
                    buttonFrame.raidIcon:Hide()
                end
            end
        
            if buttonFrame.mostTargetedIcon and buttonFrame.targetCountString then 
                if UnitIsFriend('player', buttonFrame.unitID) then
                    local foundingTargetID = getFoundingTargetID(friendlyTargets, buttonFrame.unitID)
                    if foundingTargetID then
                        if self.config.markMostTargetedFriendly and foundingTargetID == maxFriendly then
                            buttonFrame.mostTargetedIcon:SetTexture(self.config.raidIcons[4][2])
                            buttonFrame.mostTargetedIcon:Show()
                        else
                            buttonFrame.mostTargetedIcon:Hide()
                        end
                        if self.config.trackNumberTargetingFriendlies then
                            buttonFrame.targetCountString:SetText(tostring(friendlyTargets[foundingTargetID]))
                        else
                            buttonFrame.targetCountString:SetText('')
                        end
                    else
                        buttonFrame.mostTargetedIcon:Hide()
                        buttonFrame.targetCountString:SetText('')
                    end
                else
                    local foundingTargetID = getFoundingTargetID(hostileTargets, buttonFrame.unitID)
                    if self.config.markMostTargetedHostile and foundingTargetID == maxHostile then
                        buttonFrame.mostTargetedIcon:SetTexture(self.config.raidIcons[8][2])
                        buttonFrame.mostTargetedIcon:Show()
                    else
                        buttonFrame.mostTargetedIcon:Hide()
                    end
                    if self.config.trackNumberTargetingHostiles then
                        buttonFrame.targetCountString:SetText(tostring(hostileTargets[foundingTargetID]))
                    else
                        buttonFrame.targetCountString:SetText('')
                    end
                end
            end
        
        else
            buttonFrame.fontString:SetText('')
            buttonFrame.texture:SetVertexColor(unpack(self.config.missingColor))
            if buttonFrame.raidIcon then buttonFrame.raidIcon:Hide() end
            if buttonFrame.mostTargetedIcon then buttonFrame.mostTargetedIcon:Hide() end
            if buttonFrame.targetCountString then buttonFrame.targetCountString:SetText('') end
            buttonFrame.texture:SetWidth( self.buttonWidth )
        end
    end
end

--Event handlers
function targetAssist:PARTY_MEMBERS_CHANGED() self:updateRoster() end
function targetAssist:RAID_ROSTER_UPDATE() self:updateRoster() end
function targetAssist:UNIT_PET() self:updateRoster() end
function targetAssist:PLAYER_ROLES_ASSIGNED() self:updateRoster() end
function targetAssist:oRA_MainTankUpdate() self:updateRoster() end
function targetAssist:oRA3_TanksUpdated(event, tankNames) 
    self.oRA3Tanks = tankNames
    self:updateRoster()
end 
